/* ----------------------------------------------------------------------------
**      menu.c   Helper menu functions for classes/c_menu.c
** ----------------------------------------------------------------------------
**      TGT                                       TextGUIToolkit Library
**      Copyright (C)   2000-2001       Mateusz Golicz and Tomasz Sterna
**
**      LGPL license
**
** ----------------------------------------------------------------------------
**      MAINTAINER      Mateusz Golicz
**
**      Log:
**      
**
*/

#include "tgt.h"
#include <stdio.h>
#define TGT_MENU_PROTOS
#include <stdlib.h>
#define max(a,b) ((a) > (b) ? (a) : (b))

extern struct tgt_terminal * g_terminal;


void tgt_menu_rerender(tgt_cell * buffer,int xs,int ys,
		       struct tgt_int_menuitem * items,
		       int fg,int bg,int framecolor)
{
    tgt_cell attr=TGT_T_BUILDCELL(fg,bg,1,0,0);
    tgt_cell fcattr=TGT_T_FG(attr,framecolor);
    int j,i,k,n=ys-2;
    
    tgt_int_upperb(buffer,xs,TGT_T_FG(attr,framecolor));buffer+=xs;
    for(i=0;i<n;i++)
    {
	if(items[i].name)
	{
    	    *(buffer++)=TGT_T_FCHAR(fcattr,TGT_TC_VL);
    	    tgt_flprintf(buffer,xs-2,attr,"%s",items[i].name);
    	    buffer+=xs-2;
    	    *(buffer++)=TGT_T_FCHAR(fcattr,TGT_TC_VL);
	}
	else
	{
    	    *(buffer++)=TGT_T_FCHAR(fcattr,TGT_TC_TE);
    	    for(j=0,k=xs-2;j<k;j++) *(buffer++)=TGT_T_FCHAR(fcattr,TGT_TC_HL);
    	    *(buffer++)=TGT_T_FCHAR(fcattr,TGT_TC_TS);
	}
    }
    tgt_int_lowerb(buffer,xs,TGT_T_FG(attr,framecolor));buffer+=xs;
}

void tgt_menu_chcolors(tgt_cell * buffer,int n,int xs,int ys,int fg,int bg)
{
    int i,q=xs-2;
    if(n>=0 && n<=(ys-2))
    {
	buffer+=(n+1)*xs+1;
	for(i=0;i<q;i++, buffer++) *buffer=TGT_T_FG(TGT_T_BG(*buffer,bg),fg);
    }
}

void tgt_menu_enable(struct tgt_object *obj,int x,int y)
{    
    struct tgt_object *parent=obj->ln.parent;
    obj->x=x; obj->y=y;
    if(parent)
    {
	if((obj->x+obj->xs)>parent->xs) obj->x=parent->xs-obj->xs;
	if((obj->y+obj->ys)>parent->ys) obj->y=parent->ys-obj->ys;
    }
    obj->objflags&=~(TGT_OBJFLAGS_INVISIBLE);
    tgt_setfocus(obj);
    tgt_lock_mouse(obj);
}

struct tgt_object * tgt_menu_getcontext(void)
{
    return(tgt_createobject(TGT_CLASS_MENU,NULL));
}

void tgt_menu_selectability(struct tgt_object * menu,int pos,int sel)
{
    struct tgt_int_menu_exchange ex;
    ex.n = pos;
    ex.q = sel;
    menu->classf(menu,TGT_OBJECT_SETTAG,TGTA_MENU_SWITCHSELECTABILITY,(void*) &ex);
}

void tgt_menu_add(struct tgt_object * obj,char * item,int nonselectable,struct tgt_object * submenu)
{
    obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_ADDITEM,(void*) item);
    if(!item) nonselectable=1;
    if(nonselectable) obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_SWITCHSELECTABILITYLAST,0);
    if(submenu) obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_ASSIGNSUBMENU,(void*) submenu);
}

void tgt_menu_modify(struct tgt_object * obj, int num, char * text, int nonselectable, struct tgt_object * submenu)
{
    struct tgt_int_menu_exchange ex;

    ex.n = num; ex.p = text;
    obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_CHANGEITEM,(void*) &ex);
    ex.n = num; ex.p = submenu;
    obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_CHSUBMENU,(void*) &ex);
    ex.n = num; 
    if(nonselectable)
	ex.q = 0;
    else
	ex.q = 1;
    obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_SWITCHSELECTABILITY,(void*) &ex);
}

void tgt_menu_delete(struct tgt_object * obj, int n)
{
    obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_DELETEITEM,(void*) n);
}

void tgt_menu_insert(struct tgt_object * obj,int as, char * item,int nonselectable,struct tgt_object * submenu)
{
    struct tgt_int_menu_exchange ex;

    ex.n = as; ex.p = item;

    obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_INSERTITEM,(void*) &ex);
    if(!item) nonselectable=1;
    if(nonselectable)
    {
	ex.n = as;
	ex.q = 0;
	obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_SWITCHSELECTABILITY,(void*) &ex);
    }
    if(submenu) 
    {
	ex.n = as;
	ex.p = submenu;
	obj->classf(obj,TGT_OBJECT_SETTAG,TGTA_MENU_CHSUBMENU,(void*) &ex);
    }
}
